//---------------------------------------------
//  App: Summit Administration Application
//  (c) TASC Systems Inc. 2014-2015

//---------------------------------------------
// Summit Constants
//#region Summit Constants
var digitalInputsBase = 352;
var digitalInputPoints = 48;

var digitalOutputBase = 496;
var digitalOutputPoints = 16;

var analogInputPoints = 32;
var analogInputRangeDefault = "10";// Default range = 0 to 10V
var analogInputLowThreshold = "0";
var analogInputHighThreshold = "10";
var analogInputOffset = "0";
var analogOutputPoints = 4;

var SUMMIT_WEB_SERVER_PORT = 80;

var SnmpHostAddress;

var SnmpCommunity = "public"

var ERRORLOG_FAILURE = 'FAILURE';
var ERRORLOG_WARNING = 'WARNING';
var ERRORLOG_NORMAL = 'NORMAL';
//#endregion Summit Constants

//---------------------------------------------
// Initializing Logging infrastructure
//#region Logging Infrastructure
// Start the logger
var log4js = require('log4js');

//config
log4js.configure({
    appenders: [
        //{ type: 'console' },
        {
            type: "file", 
            absolute: true, 
            filename: "./log/summit.log", 
            maxLogSize: 1000000,
            backups: 10
        }
    ]
});
var logger = log4js.getLogger();
logger.setLevel(log4js.levels.INFO);// Toggle to this if you want just basic log info
//logger.setLevel(log4js.levels.DEBUG);// Toggle to this if you want all log info
//logger.setLevel(log4js.levels.ALL);// Toggle to this if you want all log info
//#endregion Logging Infrastructure

//---------------------------------------------
// Initializing server infrastructure
//#region Server Infrastructure
logger.debug('Intializing server infrastructure')

// Fire up the server + socket connections
var express = require('express');
var app = express();

// For serving static pages
var serveStatic = require('serve-static');

// For parsing middleware - e.g., JSON body
var bodyparser = require('body-parser');
app.use(bodyparser.json());
app.use(bodyparser.urlencoded({ extended: true }));

// For server and filesystem
var server = require('http').Server(app);
var fs = require('fs');
var cp = require('child_process');
var path = require('path');
var moment = require('moment');
var os = require('os');
var pjson = require('./package.json')

logger.info('============================================================================================');
logger.info(pjson.description);
logger.info('   Version ' + pjson.version + ' (' + pjson.date + ') ')
logger.info('   (c) ' + pjson.author);
logger.info('   Serial Number [coming soon: serial number of this device]');
logger.info('   ' + os.platform() + " OS Version " + os.release())
logger.info('   ' + os.freemem() + ' Free / ' + os.totalmem() + ' Total Memory');
logger.info('--------------------------------------------------------------------------------------------');

// For socket.io messaging
//#region socket.io messaging
var ClientConnectedCounter = 0;
//var io = require('socket.io')(server, { 'pingInterval': 20000, 'pingTimeout': 10000, 'transports': ['polling'], 'allowUpgrades': false });
var io = require('socket.io')(server, { 'transports': ['polling'], 'allowUpgrades': false });
io.sockets.on('connection', function (socket) {
    ClientConnectedCounter++;
    var address = socket.handshake.address;
    logger.warn('socket.connection=' + socket.id + ' client=' + address + ' (# clients=' + ClientConnectedCounter + ')');

    socket.on('reconnect', function () {
        logger.warn('socket.reconnection=' + socket.id + ' client=' + address + ' (# clients=' + ClientConnectedCounter + ')');
    });

    socket.on('disconnect', function (reason) {
        ClientConnectedCounter--;
        logger.warn('socket.disconnect=' + reason + ' id=' + socket.id + ' client=' + address + ' (# clients=' + ClientConnectedCounter + ')');
    });

    socket.on('error', function (error) {
        logger.error('socket.error=' + error + ' id=' + socket.id + ' client=' + address);
    });
});
//#endregion socket.io messaging
//#endregion Server Infrastructure

//---------------------------------------------
// Initializing Summit Users
//#region Summit Users
logger.info('Intializing Summit user management');
var summitUsers = require('./users.js');
summitUsers.Initialize(logger);
//#endregion Summit Users

//---------------------------------------------
// Initializing database infrastucture
//#region Database Infrastructure
logger.info('Intializing Summit database object')
var sqlite3 = require('sqlite3').verbose();
var db = new sqlite3.Database('data/summit.db');
//#endregion Database Infrastructure

//---------------------------------------------
// Initialize System Status Manager
//#region System Status Manager
logger.info('Intializing Summit system status manager')
var systemStatusManager = require('./systemstatus.js')
systemStatusManager.InitializeSystemStatusManager(db, logger, function () {
    systemStatusManager.LogSystemStatus(pjson.description + " Version " + pjson.version + " started", ERRORLOG_NORMAL , fs, io, logger, db);
    systemStatusManager.EnsureFactoryResetInterfacesFile(logger);
});
//#endregion System Status Manager

//---------------------------------------------
// Initializing network management
//#region Network Management
logger.info('Intializing Summit Network and SNMP objects')
var networkManager = require('./networkmanager.js');
var netSnmp = require('net-snmp');
var nodemailer = require('nodemailer');// SMTP support
networkManager.InitializeNetworkInterfaces(db, logger, netSnmp);

var smtpManager = require('./smtp.js');
smtpManager.InitializeSmtp(nodemailer, logger, function (smtp) {
    // Do nothing for now ... maybe later, maybe never
});
//#endregion Network Management

//---------------------------------------------
// Initializing Kenwood Manager
//#region Kenwood Manager
logger.info('Intializing Kenwood Manager objects')
var kenwoodManager = require('./kenwood.js')
var KENWOOD_PING_INTERVAL = 10// Default = 10seconds
//#endregion Kenwood Manager

//---------------------------------------------
// Initializing IO infrastucture
//#region IO Infrastructure
logger.info('Intializing I/O')
var WAIT_UNTIL_SYSTEM_STARTS = 500

var DIGITAL_INPUTS_CYCLE_TIME = 500// Default = 400ms/500ms
var DIGITAL_OUTPUTS_CYCLE_TIME = 500// Default = 400ms
var ANALOG_INPUTS_CYCLE_TIME = 750// Default = 500ms/750ms
var ANALOG_OUTPUTS_CYCLE_TIME = 500// Default = 200ms

var digitalInputs = require('./digitalinputs.js');
var digitalOutputs = require('./digitaloutputs.js');
var analogInputs = require('./analoginputs.js');
var analogOutputs = require('./analogoutputs.js');
var SerialPortsManager = require('./serialports.js');
//#endregion IO Infrastructure

//---------------------------------------------
// Initializing Factory Reset
//#region Factory Reset
logger.info('Intializing Factory Reset')

var factoryReset = require('./factoryreset.js');
//#endregion Factory Reset

//---------------------------------------------
// Initializing API Middleware
//#region API Middleware
logger.info('Initializing Summit API')
app.use(serveStatic('.', { 'index': ['index.html'] }))
app.use(serveStatic('views', { 'index': false }))
//#endregion API Middleware 

//---------------------------------------------
// Initializing API Utility Functions
//#region API Utility Functions
app.get('/api/causeerror', function (req, res, next) {
    logger.info('Summit API: GET /api/causeerror');
    res.send('caused error');
    process.exit(-1);
})
//#endregion API Utility Functions

//#region Summit User

// Login Summit User
app.put('/api/login', function (req, res, next) {
    logger.info('Summit API: PUT /api/login');
    var ip = req.headers['x-real-ip'] || req.connection.remoteAddress
    summitUsers.LoginUser(req.body, ip, function (success) {
        if (success) {
            logger.info('Summit API: Login attempt by user=' + req.body.user + " was successful");
            res.send({ "LoginPassed": true });
        }        
        else {
            logger.info('Summit API: Login attempt by user=' + req.body.user + " failed");
            res.send({ "LoginPassed": false });
        }
    });
});

// Get Summit Login Attempts
app.get('/api/login/attempts', function (req, res, next) {
    logger.info('Summit API: GET /api/login/attempts');
    summitUsers.GetSummitLoginAttempts(function (SummitLoginAttempts) {
        logger.info("Summit API: Summit Login Attempts=" + SummitLoginAttempts.length);
        res.send(SummitLoginAttempts);
    });
});

// Get Summit User
app.get('/api/login/users', function (req, res, next) {
    logger.info('Summit API: GET /api/login/users');
    summitUsers.GetSummitUser(function (SummitUser) {
        logger.info("Summit API: Summit User=" + SummitUser.user);
        res.send(SummitUser);
    });
});

// Put Summit User
app.put('/api/login/users', function (req, res, next) {
    logger.info('Summit API: PUT /api/login/users');
    summitUsers.PutSummitUser(req.body, function (SummitUser) {
        logger.info("Summit API: Summit User=" + SummitUser.user);
        res.send(SummitUser);
    });
});

//#endregion Summit User

//#region Summit System API
//---------------------------------------------
// Summit System API

// Identify system
app.get('/api/summit/identify', function (req, res, next) {
    logger.info('Summit API: GET /api/summit/identify');
    // Flash lights yellow 1s, then blue 1s for a total of 1 minute
    var timeDelay = 3000;
    for (var i = 0; i < 5; i++) {
        setTimeout(function () {
            systemStatusManager.WarningOperation(fs, io, logger);
        }, timeDelay * i);
        
        setTimeout(function () {
            systemStatusManager.NormalOperation(fs, io, logger);
        }, timeDelay * i + (timeDelay / 2));
    }
    res.send('Identify Device OK');
})

// Get installed system software
app.get('/api/summit/operatingsystem', function (req, res, next) {
    logger.info('Summit API: GET /api/summit/operatingsystem');
    data = { "os": os.platform() + " OS Version " + os.release() };
    logger.debug('Summit API: GET /api/summit/operatingsystem - System OS=' + JSON.stringify(data))
    res.send(data);
})

// Put System Date/Time
app.put('/api/summit/systemdatetime', function (req, res, next) {
    logger.info('Summit API: PUT /api/summit/systemdatetime ' + JSON.stringify(req.body));
    var exec = cp.exec;
    var ret;
    exec('date -s "' + req.body.SystemDateTime + '"', function (error, stdout, stderr) {
        if (error) res.send({ "data": "ERROR: " + error });
        else {
            res.send({ "data": "OK: "+ stdout });
        }
    });
})

// Get current system status
app.get('/api/summit/systemstatus', function (req, res, next) {
    logger.info('Summit API: GET /api/summit/systemstatus');
    var CurrentSystemStatus = systemStatusManager.GetCurrentSystemStatus(logger);
    logger.debug('Summit API: GET /api/summit/systemstatus - Current System Status=' + CurrentSystemStatus);
    res.send(CurrentSystemStatus);
})

// Get list of system events
app.get('/api/summit/systemevents', function (req, res, next) {
    logger.info('Summit API: GET /api/summit/systemstatus');
    var systemStatus = [];
    db.all("SELECT ErrorMessage, ErrorType, DATE(TimeChanged) FROM ErrorLog ORDER BY DATE(TimeChanged) DESC LIMIT 20", function (err, rows) {
        if (err) {
            throw logger.error("Summit API: GET /api/summit/systemstatus - ERROR=" + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            systemStatus.push({
                "ErrorType"     : rows[i].ErrorType,
                "ErrorMessage"  : rows[i].ErrorMessage,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.debug('Summit API: GET /api/summit/systemstatus - System Status=' + JSON.stringify(systemStatus));
        res.json(systemStatus);
    });
})
//#endregion

//#region Crest Software
// Get Crest software info
app.get('/api/crest', function (req, res, next) {
    logger.info('Summit API: GET /api/crest');
    logger.debug('Summit API: GET /api/crest - Crest info=' + JSON.stringify(pjson))
    res.send(pjson);
})

app.get('/api/crest/latest', function (req, res, next) {
    logger.info('Summit API: GET /api/crest/latest');
    systemStatusManager.GetLatestCrestVersionFromServer(logger, function (latestversion) {
        logger.debug('Summit API: GET /api/crest/latest - Crest latest version=' + latestversion)
        res.send(latestversion);
    });
})

app.put('/api/crest/download', function (req, res, next) {
    logger.info('Summit API: GET /api/crest/download');
    logger.info('Summit API: GET /api/crest/download ' + JSON.stringify(req.body));
    systemStatusManager.DownloadCrest(req.body.CrestUrl,logger, function (result) {
        logger.debug('Summit API: GET /api/crest/download - Crest download=' + result);
        res.send(result);
    });
    // This is here to imitate the upgrade process for testing purposes only
    //setTimeout(function () {
    //    res.send("Download OK");
    //}, 1500);
})

//#endregion Crest Software

//#region Networking, SMTP and SNMP API
//---------------------------------------------
// Networking, SMTP and SNMP

// Networking Support
//#region Networking Support
// Get network interfaces
app.get('/api/network/interfaces', function (req, res, next) {
    
    // Get Network Interfaces information from the Network Manager module
    logger.info('Summit API: GET /api/network/interfaces - Get Network Interfaces called');
    networkManager.GetNetworkInterfaces(function (err, networkInterfaces) {
        if (err) {
            logger.info('Summit API: GET /api/network/interfaces - Return Network Interfaces. Error=' + err);
            res.json({});
        } else {
            // Return the network interfaces
            logger.info('Summit API: GET /api/network/interfaces - Return Network Interfaces=' + JSON.stringify(networkInterfaces));
            res.json(networkInterfaces);        
        }
    });

})

// Set network interfaces
app.put('/api/network/interfaces', function (req, res, next) {
    logger.info('Summit API: PUT /api/network/interfaces - Updating Network Interface=' + JSON.stringify(req.body));
    networkManager.PutNetworkInterfaces(req.body, function (error, interfaces) {
        if (error) {
            logger.error('Summit API: PUT /api/network/interfaces - Updated Network Interface. Error=' + error);
        } else {
            logger.info('Summit API: PUT /api/network/interfaces - Updated Network Interface=' + JSON.stringify(interfaces));
        }
        res.send(interfaces);
    });
})
//#endregion Networking Support

// SMTP Support
//#region SMTP Support

// Get SMTP Settings
app.get('/api/network/smtp', function (req, res, next) {
    logger.info('Summit API: GET /api/network/smtp');
    smtpManager.GetSmtpSettings(function (smtp){
        logger.info('Summit API: GET /api/network/smtp - SMTP Settings=' + smtp);
        res.json(smtp);    
    })
});

// Put SMTP Settings
app.put('/api/network/smtp', function (req, res, next) {
    logger.info('Summit API: PUT /api/network/smtp');
    req.body.Default = false;// No longer using default value
    smtpManager.SetSmtpSettings(req.body, function (smtp) {
        logger.info('Summit API: PUT /api/network/smtp - SMTP Settings=' + smtp);
        res.json(smtp);
    })
});

// Test a SMTP host 
app.put('/api/network/smtp/test', function (req, res, next) {
    logger.info('Summit API: PUT /api/network/smtp/test - data='+JSON.stringify(req.body));
    
    // create reusable transporter object using the default SMTP transport
    var transporter = nodemailer.createTransport({
        port: req.body.Port,
        host: req.body.Host,
        secure: req.body.Secure,
        auth: {
            user: req.body.User,
            pass: req.body.Pass
        }
    });
    
    // setup e-mail data with unicode symbols
    var mailOptions = {
        from: req.body.From, // sender address
        to: req.body.To, // list of receivers
        subject: 'Crest SMTP Test Email', // Subject line
        text: 'Your SMTP configuration is OK and so Crest is able to send you emails!', // plaintext body
    };
    
    // send mail with defined transport object
    var result = false;
    transporter.sendMail(mailOptions, function (error, info) {
        if (error) {
            logger.error("Summit API: PUT /api/network/smtp/test - send email error=" + error);
            result = false;
            res.json({ "result": result , "response": error });
        } else {
            logger.info("Summit API: PUT /api/network/smtp/test - send email result=" + info.response);
            result = true;       
            res.json({ "result": result, "response": info.response });
        }
    });    
});

//#endregion SMTP Support

// SNMP Support
//#region SNMP Support
// Get SNMP hosts
app.get('/api/network/snmp/hosts', function (req, res, next) {
    logger.info('Summit API: GET /api/network/snmp/hosts');
    
    // Get the complete list of SNMP Hosts and update I/O processing sub-systems
    var snmpHosts = [];
    db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
        if (err) {
            // LogSystemStatus(errorMessage, errorType, timeStamp, db, systemStatusManager)
            throw logger.error('Summit API: GET /api/network/snmp/hosts - Error=' + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            snmpHosts.push({
                "Id"            : rows[i].Id,
                "HostName"      : rows[i].HostName,
                "Enabled"       : rows[i].Enabled,
                "IpAddress"     : rows[i].IpAddress,
                "Port"          : rows[i].Port,
                "Community"     : rows[i].Community
            });
        }
        logger.debug('Summit API: GET /api/network/snmp/hosts - Hosts=' + JSON.stringify(snmpHosts));
        
        // (Re)set the SNMP Host Sessions for I/O Processors
        networkManager.CreateSnmpHostSessions(snmpHosts, function (sessions) {
            digitalInputs.SetSnmpHostSessionsArray(sessions, logger);
            digitalOutputs.SetSnmpHostSessionsArray(sessions, logger);
            analogInputs.SetSnmpHostSessionsArray(sessions, logger);
            analogOutputs.SetSnmpHostSessionsArray(sessions, logger);
        });
        res.json(snmpHosts);
    });
})

// Delete a SNMP host
app.put('/api/network/snmp/hosts/delete', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  HostName    (TEXT)      <host name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  IpAddress   (TEXT)      <IP Address of Host>
    //  Port        (TEXT)      <Port for host, typically 162>
    //  Community   (TEXT)      <SNMP community name>
    logger.info('Summit API: PUT /api/network/snmp/hosts/delete');
    
    // Update the existing record
    logger.debug('Summit API: PUT /api/network/snmp/hosts/delete - Host to delete=' + JSON.stringify(req.body));
    
    var update = "DELETE FROM SnmpHosts WHERE rowid='" + req.body.Id + "'";
    db.run(update, function (err) {
        if (err !== null) {
            logger.error('Summit API: PUT /api/network/snmp/hosts/delete - Error (Update)=' + JSON.stringify(err));
            res.send({ msg: "SQLite3 dB: api/network/snmp/hosts/" + " DELETE FROM  err=" + JSON.stringify(err) })
        }
        
        // Get the complete list of SNMP Hosts and update I/O processing sub-systems
        var snmpHosts = [];
        db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
            if (err) {
                throw logger.error('Summit API: PUT /api/network/snmp/hosts/delete - Error (Read)=' + JSON.stringify(err));
            }
            for (i = 0; i < rows.length; i++) {
                // Always send the raw value - clients responsible for conversion
                snmpHosts.push({
                    "Id"            : rows[i].Id,
                    "HostName"      : rows[i].HostName,
                    "Enabled"       : rows[i].Enabled,
                    "IpAddress"     : rows[i].IpAddress,
                    "Port"          : rows[i].Port,
                    "Community"     : rows[i].Community
                });
            }
            logger.debug('Summit API: PUT /api/network/snmp/hosts/delete - Hosts=' + JSON.stringify(snmpHosts));           
            res.json(snmpHosts);
        });
    });
})

// Create new SNMP host
app.put('/api/network/snmp/hosts', function (req, res, next) {
    //  HostName    (TEXT)      <host name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  IpAddress   (TEXT)      <IP Address of Host>
    //  Port        (TEXT)      <Port for host, typically 162>
    //  Community   (TEXT)      <SNMP community name>
    logger.info('Summit API: PUT /api/network/snmp/hosts');
    
    // Add a new record
    logger.debug('Summit API: PUT /api/network/snmp/hosts - Host to add=' + JSON.stringify(req.body));
    
    // Insert the new record
    db.run("BEGIN")
    insertStatement = "'" + req.body.HostName + "', '" + req.body.Enabled + "', '" + req.body.IpAddress + "', '" + req.body.Port + "', '" + req.body.Community + "'";
    db.run("INSERT INTO SnmpHosts (HostName, Enabled, IpAddress, Port, Community) VALUES (" + insertStatement + ")", function (err) {
        if (err !== null) {
            logger.error('Summit API: PUT /api/network/snmp/hosts - Error (Insert)=' + err);
            res.send({ msg: "SQLite3 dB: api/snmphosts Insert err=" + JSON.stringify(err) })
        }
        
        // Successful insert, just look up the ID for future reference
        logger.debug('Summit API: PUT /api/network/snmp/hosts - ID=' + this.lastID);
        req.body.Id = this.lastID;
    });
    db.run("COMMIT");
    
    // Get the complete list of SNMP Hosts and update I/O processing sub-systems
    var snmpHosts = [];
    db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
        if (err) {
            // LogSystemStatus(errorMessage, errorType, timeStamp, db, systemStatusManager)
            throw logger.error('Summit API: PUT /api/network/snmp/hosts - Error (Read)=' + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            snmpHosts.push({
                "Id"            : rows[i].Id,
                "HostName"      : rows[i].HostName,
                "Enabled"       : rows[i].Enabled,
                "IpAddress"     : rows[i].IpAddress,
                "Port"          : rows[i].Port,
                "Community"     : rows[i].Community
            });
        }
        logger.debug('Summit API: PUT /api/network/snmp/hosts - Hosts=' + JSON.stringify(snmpHosts));
        
        // (Re)set the SNMP Host Sessions for I/O Processors
        networkManager.CreateSnmpHostSessions(snmpHosts, function (sessions) {
            digitalInputs.SetSnmpHostSessionsArray(sessions, logger);
            digitalOutputs.SetSnmpHostSessionsArray(sessions, logger);
            analogInputs.SetSnmpHostSessionsArray(sessions, logger);
            analogOutputs.SetSnmpHostSessionsArray(sessions, logger);
        });
        res.json(snmpHosts);
    });
})

// Modify existing SNMP host
app.put('/api/network/snmp/hosts/:snmphost', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  HostName    (TEXT)      <host name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  IpAddress   (TEXT)      <IP Address of Host>
    //  Port        (TEXT)      <Port for host, typically 162>
    //  Community   (TEXT)      <SNMP community name>
    logger.info('Summit API: PUT /api/network/snmp/hosts/' + req.params.snmphost);
    
    // Update the existing record
    var update = "UPDATE SnmpHosts SET HostName='" + req.body.HostName + "', Enabled='" + req.body.Enabled + "', IpAddress='" + req.body.IpAddress + "', Port='" + req.body.Port + "', Community='" + req.body.Community + "' WHERE rowid='" + req.params.snmphost + "'";
    db.run(update, function (err) {
        if (err !== null) {
            throw logger.error('Summit API: PUT /api/network/snmp/hosts/' + req.params.snmphost + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ msg: "SQLite3 dB: api/snmphosts/" + req.body.Id + " UPDATE err=" + JSON.stringify(err) })
        }
    });
    
    // Get the complete list of SNMP Hosts and update I/O processing sub-systems
    var snmpHosts = [];
    db.all("SELECT rowid as Id, HostName, Enabled, IpAddress, Port, Community FROM snmphosts ORDER BY Id", function (err, rows) {
        if (err) {
            throw logger.error('Summit API: PUT /api/network/snmp/hosts/' + req.params.snmphost + ' - Error (Read)=' + err);
        }
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            snmpHosts.push({
                "Id"            : rows[i].Id,
                "HostName"      : rows[i].HostName,
                "Enabled"       : rows[i].Enabled,
                "IpAddress"     : rows[i].IpAddress,
                "Port"          : rows[i].Port,
                "Community"     : rows[i].Community
            });
            
            if (rows[i].Id == req.params.snmphost) {
                snmpHostReturn = rows[i];
            }
        }
        logger.debug('Summit API: PUT /api/network/snmp/hosts/' + req.params.snmphost + ' - Hosts=' + JSON.stringify(snmpHosts));
        
        // (Re)set the SNMP Host Sessions for I/O Processors
        networkManager.CreateSnmpHostSessions(snmpHosts, function (sessions) {
            digitalInputs.SetSnmpHostSessionsArray(sessions, logger);
            digitalOutputs.SetSnmpHostSessionsArray(sessions, logger);
            analogInputs.SetSnmpHostSessionsArray(sessions, logger);
            analogOutputs.SetSnmpHostSessionsArray(sessions, logger);
        });
        res.send(snmpHostReturn);
    });
})

// Test a SNMP host by sending a trap message
app.put('/api/network/snmp/test', function (req, res, next) {
    logger.info('Summit API: PUT /api/network/snmp/test');
    
    logger.debug('Summit API: PUT /api/network/snmp/test - Host=' + JSON.stringify(req.body));
    // Create session for test send
    var SnmpSessionOptions = {
        port: 161,
        retries: 1,
        timeout: 5000,
        transport: "udp4",
        trapPort: 162,
        version: netSnmp.Version2c
    };
    SnmpHostAddress = req.body.IpAddress;
    SnmpCommunity = req.body.Community;
    SnmpOptions = JSON.parse(JSON.stringify(SnmpSessionOptions));
    SnmpOptions.trapPort = req.body.Port;
    TestSession = netSnmp.createSession(SnmpHostAddress, SnmpCommunity, SnmpOptions);
    
    // Build the trap
    var trapOid = "1.3.6.1.4.1.11902.10.5.1";
    var varbinds = [
        {
            oid: "1.3.6.1.4.1.11902.10.5.2",
            type: netSnmp.ObjectType.OctetString,
            value: "Test Trap"
        },
    ];
    
    // Send the trap
    TestSession.trap(trapOid, varbinds, function (error) {
        if (error)
            logger.error('Summit API: PUT /api/network/snmp/test - Error (Trap send)=' + error);
        else
            logger.debug('Summit API: PUT /api/network/snmp/test - Success');
    });
    res.send(req.body);
});

// Return the TASC-SUMMIT.MIB file
app.get('/api/network/snmp/mibfile', function (req, res, next) {
    logger.info('Summit API: GET /api/network/snmp/mibfile');
    
    // Read the TASC-SUMMIT.MIB file
    fs.readFile("./TASC-SUMMIT.MIB", 'utf8', function (err, data) {
        if (err) {
            logger.error('Summit API: GET /api/network/snmp/mibfile - Error (Read)=' + JSON.stringify(err));
            res.send("Error reading TASC-SUMMIT.MIB file");
        }
        else {
            logger.debug('Summit API: GET /api/network/snmp/mibfile - File=TASC-SUMMIT.MIB');
            res.send(encodeURIComponent(data));
        }
    });
})
//#endregion SNMP Support

//#endregion Networking, SMTP and SNMP API

//#region Logging API
//---------------------------------------------
// Logging API

app.get('/api/logging/loglevel', function (req, res, next) {
    logger.info('Summit API: GET /api/logging/loglevel');
    
    var loglevel;
    switch (logger.level.levelStr) {
        case "INFO":
            loglevel = "BASIC";
            break;

        case "DEBUG":
            loglevel = "VERBOSE";
            break;

        case "ALL":
            loglevel = "ALL";
            break;

    }
    logger.info('Summit API: PUT /api/logging/loglevel - Level=' + loglevel);
    res.json({ "LogLevel": loglevel });
});

app.put('/api/logging/loglevel/:loglevel', function (req, res, next) {
    logger.info('Summit API: PUT /api/logging/loglevel/' + req.params.loglevel);
    
    switch (req.params.loglevel) {
        case "BASIC":
            logger.setLevel(log4js.levels.INFO);
            loglevel = "BASIC";
            logger.info("Logging Manager - Level=" + loglevel + " (" + logger.level.levelStr + ")");
            break;

        case "VERBOSE":
            logger.setLevel(log4js.levels.DEBUG);
            loglevel = "VERBOSE";
            logger.info("Logging Manager - Level=" + loglevel + " (" + logger.level.levelStr + ")");
            break;

        case "ALL":
            logger.setLevel(log4js.levels.ALL);
            loglevel = "ALL";
            logger.info("Logging Manager - Level=" + loglevel + " (" + logger.level.levelStr + ")");
            break;
    }
    logger.info('Summit API: PUT /api/logging/loglevel/' + req.params.loglevel + " - Level=" + logger.level.levelStr);
    res.json({ "LogLevel": loglevel });
});

// Get list of log files
app.get('/api/logging/logfiles', function (req, res, next) {
    logger.info('Summit API: GET /api/logging/logfiles');
    
    // Read the log files
    logDirectory = './log';
    var logFiles = [];
    fs.readdir(logDirectory, function (err, files) {
        if (err) {
            logger.error('Summit API: GET /api/logging/logfiles - Error (Read)=' + JSON.stringify(err));
            res.send('Error reading log files');
        }
        else {
            // Get a list of these files
            for (var i in files) {
                var filePath = logDirectory + '/' + files[i];
                var stat = fs.statSync(filePath);
                if (stat.isFile()) {
                    logFiles.push({
                        "Name"  : files[i],
                        "Size"  : parseInt(stat.size).toLocaleString(),
                        "Date"  : stat.mtime
                    });
                }
            }
            
            // Put the files in the order that we want
            logFiles.sort().reverse(); // this will put 'summit.log' at end, but sort with most recent at front 
            var summitLog = logFiles.pop();// get 'summit.log'
            logFiles.unshift(summitLog);// put 'summit.log' in front
            
            logger.debug('Summit API: GET /api/logging/logfiles - Files=' + JSON.stringify(logFiles));
            res.send(logFiles);
        }
    });
});

// Read contents of a log file
app.get('/api/logging/logfiles/:logfile', function (req, res, next) {
    logger.info('Summit API: GET /api/logging/logfiles/' + req.params.logfile);
    
    // Read the log files
    logDirectory = './log';
    fs.readFile(logDirectory + "/" + req.params.logfile, 'utf8', function (err, data) {
        if (err) {
            logger.error('Summit API: GET /api/logging/logfiles/' + req.params.logfile + ' - Error (Read)=' + JSON.stringify(err));
            res.send("Error reading log file");
        }
        else {
            logger.debug('Summit API: GET /api/logging/logfiles - File=' + req.params.logfile);
            res.send(encodeURIComponent(data));
        }
    });
})
//#endregion Logging API

//#region Digital Inputs

// Get Digital Input points
app.get('/api/digitalinputs', function (req, res, next) {
    logger.info('Summit API: GET /api/digitalinputs');
    
    var _digitalInputs = [];
    db.all("SELECT Address, Name, Value, Enabled, SendEvents, Polarity, HoldTime, TimeChanged FROM digitalinputs ORDER BY Address", function (err, rows) {
        if (err) {
            // LogSystemStatus(errorMessage, errorType, timeStamp, db, systemStatusManager)
            throw logger.error('Summit API: GET /api/digitalinputs - Error (Select)=' + JSON.stringify(err));
        }
        
        for (i = 0; i < rows.length; i++) {
            // Always send the raw value - clients responsible for conversion
            _digitalInputs.push({
                "Address"       : rows[i].Address,
                "Name"          : rows[i].Name,
                "Value"         : rows[i].Value,
                "Enabled"       : rows[i].Enabled,
                "SendEvents"    : rows[i].SendEvents,
                "Polarity"      : rows[i].Polarity,
                "HoldTime"      : rows[i].HoldTime,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.debug('Summit API: GET /api/digitalinputs - Points=' + JSON.stringify(_digitalInputs));
        res.json(_digitalInputs);
    });
})

// Get Digital Input point
app.get('/api/digitalinputs/:digitalinput', function (req, res, next) {
    logger.info('Summit API: GET /api/digitalinputs/' + req.params.digitalinput);
    
    var _digitalInput = [];
    db.get("SELECT Address, Name, Value, Enabled, SendEvents, Polarity, HoldTime, TimeChanged FROM digitalinputs WHERE Address='" + req.params.digitalinput + "'", function (err, row) {
        if (err) {
            throw logger.error('Summit API: GET /api/digitalinputs - Error (Select)=' + JSON.stringify(err));
        }
        
        if (typeof row !== "undefined") {
            logger.info("RESTful GET  : Digital Inputs address=" + row.Address);
            // Always send the raw value - clients responsible for conversion
            _digitalInput.push({
                "Address"       : row.Address,
                "Name"          : row.Name,
                "Value"         : row.Value,
                "Enabled"       : row.Enabled,
                "SendEvents"    : row.SendEvents,
                "Polarity"      : row.Polarity,
                "HoldTime"      : row.HoldTime,
                "TimeChanged"   : row.TimeChanged
            });
            logger.info("RESTful GET  : Digital Input - Address=" + row.Address);
            res.json(_digitalInput);
        }
        else {
            throw logger.error('Summit API: GET /api/digitalinputs/' + req.params.digitalinput + ' - Error (Select)=' + JSON.stringify(err));
            res.json({ Error: "Undefined digitalinput" });
        }
    });
})

// Update Digital Input point
app.put('/api/digitalinputs/:digitalinput', function (req, res, next) {
    //  Name        (TEXT)      <device name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  SendEvents  (TEXT)      ”TRUE”, “FALSE”
    //  Polarity    (TEXT)      ”Normally Closed”, “Normally Open”
    //  HoldTime    (INTEGER)   <0-255 seconds>
    logger.info('Summit API: PUT /api/digitalinputs/' + req.params.digitalinput);
    
    var update = "UPDATE digitalinputs SET Name='" + req.body.Name + "', Enabled='" + req.body.Enabled + "', SendEvents='" + req.body.SendEvents + "', Polarity='" + req.body.Polarity + "', HoldTime='" + req.body.HoldTime + "' WHERE address='" + req.params.digitalinput + "'";
    db.run(update, function (err) {
        if (err !== null) {
            logger.error('Summit API: PUT /api/digitalinputs/' + req.params.digitalinput + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ Error : 'Error updating digitalinput' })
        }
        else {
            // Modify the memory array
            digitalInputs.UpdateDigitalInputConfiguration(req.params.digitalinput, req.body.Name, req.body.Value, req.body.Enabled, req.body.SendEvents, req.body.Polarity, req.body.HoldTime, req.body.TimeChanged, io, logger);
            
            // Send back the req as results if everything OK
            logger.debug('Summit API: PUT /api/digitalinputs/' + req.params.digitalinput + ' - Point=' + JSON.stringify(req.body));
            res.send(JSON.stringify(req.body))
        }
    });
})

//#endregion Digital Inputs

//#region Digital Outputs

// Get Digital Ouput points
app.get('/api/digitaloutputs', function (req, res, next) {
    logger.info('Summit API: GET /api/digitaloutputs');
    
    var _digitalOutputs = [];
    db.all("SELECT Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged FROM digitaloutputs ORDER BY Address", function (err, rows) {
        if (err) {
            throw logger.error('Summit API: GET /api/digitaloutputs - Error (Select)=' + JSON.stringify(err));
        }
        
        // Always send the raw value - clients responsible for conversion
        for (i = 0; i < rows.length; i++) {
            _digitalOutputs.push({
                "Address"       : rows[i].Address,
                "Name"          : rows[i].Name,
                "Value"         : rows[i].Value,
                "Enabled"       : rows[i].Enabled,
                "SendEvents"    : rows[i].SendEvents,
                "Polarity"      : rows[i].Polarity,
                "ActionType"    : rows[i].ActionType,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.debug('Summit API: GET /api/digitaloutputs - Points=' + JSON.stringify(_digitalOutputs));
        res.json(_digitalOutputs);
    });
})

// Get Digital Output point
app.get('/api/digitaloutputs/:digitaloutput', function (req, res, next) {
    logger.info('Summit API: GET /api/digitaloutputs/' + req.params.digitaloutput);
    
    var _digitalOutput = [];
    db.get("SELECT Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged FROM digitaloutputs WHERE Address='" + req.params.digitaloutput + "'", function (err, row) {
        if (err) {
            throw logger.error('Summit API: GET /api/digitaloutputs/' + req.params.digitaloutput + ' - Error (Select)=' + JSON.stringify(err));
        }
        
        if (typeof row !== "undefined") {
            // Always send the raw value - clients responsible for conversion
            _digitalOutput.push({
                "Address"       : row.Address,
                "Name"          : row.Name,
                "Value"         : row.Value,
                "Enabled"       : row.Enabled,
                "SendEvents"    : row.SendEvents,
                "Polarity"      : row.Polarity,
                "ActionType"    : row.ActionType,
                "TimeChanged"   : row.TimeChanged
            });
            logger.debug('Summit API: GET /api/digitaloutputs/' + req.params.digitaloutput + ' - Point=' + JSON.stringify(_digitalOutput));
            res.json(_digitalOutput);
        }
        else {
            logger.error('Summit API: GET /api/digitaloutputs/' + req.params.digitaloutput + ' - Error (Select)=' + JSON.stringify(err));
            res.json({ Error: "Undefined digitaloutput" });
        }
    });
})

// Update Digital Output point
app.put('/api/digitaloutputs/:digitaloutput', function (req, res, next) {
    //  Name        (TEXT)      <device name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  SendEvents  (TEXT)      ”TRUE”, “FALSE”
    //  Polarity    (TEXT)      ”Normally Closed”, “Normally Open”
    //  ActionType  (TEXT)      "Latched", "Pulse"
    logger.info('Summit API: PUT /api/digitaloutputs/' + req.params.digitaloutput);
    
    var update = "UPDATE digitaloutputs SET Name='" + req.body.Name + "', Enabled='" + req.body.Enabled + "', SendEvents='" + req.body.SendEvents + "', Polarity='" + req.body.Polarity + "', ActionType='" + req.body.ActionType + "' WHERE address='" + req.params.digitaloutput + "'";
    db.run(update, function (err) {
        if (err !== null) {
            logger.error('Summit API: PUT /api/digitaloutputs/' + req.params.digitaloutput + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ Error : 'Error updating digitaloutput' })
        }
        else {
            // Modify the memory array
            digitalOutputs.UpdateDigitalOutputConfiguration(req.params.digitaloutput, req.body.Name, req.body.Value, req.body.Enabled, req.body.SendEvents, req.body.Polarity, req.body.ActionType, req.body.TimeChanged, io, logger)
            
            // Send back the req as results if everything OK
            logger.debug('Summit API: PUT /api/digitaloutputs/' + req.params.digitaloutput + ' - Point=' + JSON.stringify(req.body));
            res.send(JSON.stringify(req.body))
        }
    });
})

// Set Digital Output point value
app.put('/api/digitaloutputs/:digitaloutput/:value', function (req, res, next) {
    logger.info('Summit API: PUT /api/digitaloutputs/' + req.params.digitaloutput + '/' + req.params.value);
    
    // Get the digital output settings to determine if this is NC or NO
    db.get("SELECT Polarity FROM digitaloutputs WHERE Address='" + req.params.digitaloutput + "'", function (err, row) {
        if (typeof row !== "undefined") {
            // Ensure we know the proper polarity
            polarity = row.Polarity;
        }
        
        // Determine the value to set
        var index = digitalOutputBase + parseInt(req.params.digitaloutput);
        if (polarity === "Normally Open") {
            if (req.params.value == "0") {
                value = "0";
            }
            else {
                value = "1";
            }
        } 
        else {
            if (req.params.value == "0") {
                value = "1";
            }
            else {
                value = "0";
            }
        }
        
        // Set the value ...
        var index = digitalOutputBase + parseInt(req.params.digitaloutput);
        fs.writeFile("/sys/class/gpio/gpio" + index + "/value", value, function (err) {
            if (err) {
                logger.error('Summit API: PUT /api/digitaloutputs/' + req.params.digitaloutput + '/' + req.params.value + ' - Error (Write)=' + JSON.stringify(err));
                res.send({ Error : 'Error setting digitaloutput' })
            }
            
            logger.debug('Summit API: PUT /api/digitaloutputs/' + req.params.digitaloutput + '/' + req.params.value + ' - Point=' + JSON.stringify(req.params));
            res.send(req.params)
        });
    });
})

//#endregion Digital Outputs

//#region Analog Inputs

// Get Analog Input points
app.get('/api/analoginputs', function (req, res, next) {
    logger.info('Summit API: GET /api/analoginputs');
    
    var _analogInputs = [];
    db.all("SELECT Address, Name, FormulaId, FormulaEnabled, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged FROM analoginputs ORDER BY Address", function (err, rows) {
        if (err) {
            throw logger.error('Summit API: GET /api/analoginputs - Error (Select)=' + JSON.stringify(err));
        }
        
        for (i = 0; i < rows.length; i++) {
            _analogInputs.push({
                "Address"       : rows[i].Address,
                "Name"          : rows[i].Name,
                "FormulaId"     : rows[i].FormulaId,
                "FormulaEnabled": rows[i].FormulaEnabled,
                "RawValue"      : rows[i].RawValue,
                "FormulaValue"  : analogInputs.getFormulaValue(i, logger),
                "Units"         : analogInputs.getAnalogInputUnits(i,logger),
                "TempValue"     : rows[i].RawValue,
                "RangedValue"   : rows[i].RangedValue,
                "ThresholdState": rows[i].ThresholdState,
                "Enabled"       : rows[i].Enabled,
                "SendEvents"    : rows[i].SendEvents,
                "Range"         : rows[i].Range,
                "LowThreshold"  : rows[i].LowThreshold,
                "HighThreshold" : rows[i].HighThreshold,
                "FormulaLowThreshold"  : analogInputs.getFormulaLowThreshold(i, logger),
                "FormulaHighThreshold" : analogInputs.getFormulaHighThreshold(i, logger),
                "Offset"        : rows[i].Offset,
                "HoldTime"      : rows[i].HoldTime,
                "Qualifier"     : rows[i].Qualifier,
                "QualifierValue": analogInputs.getAnalogInputQualifierValue(i, logger),
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.debug('Summit API: GET /api/analoginputs - Points=' + JSON.stringify(_analogInputs));
        res.json(_analogInputs);
    });
})

// Get Analog Input point
app.get('/api/analoginputs/:analoginput', function (req, res, next) {
    logger.info('Summit API: GET /api/analoginputs/' + req.params.analoginput);
    
    var _analogInput = [];
    db.get("SELECT Address, Name, FormulaId, FormulaEnabled, RawValue, RangedValue, ThresholdState, Enabled, SendEvents, Range, LowThreshold, HighThreshold, Offset, HoldTime, Qualifier, TimeChanged FROM analoginputs WHERE Address='" + req.params.analoginput + "'", function (err, row) {
        if (err) {
            throw logger.error('Summit API: GET /api/analoginputs/' + req.params.analoginput + ' - Error (Select)=' + JSON.stringify(err));
        }
        
        if (typeof row !== "undefined") {
            _analogInput.push({
                "Address"       : row.Address,
                "Name"          : row.Name,
                "FormulaId"     : row.FormulaId,
                "FormulaEnabled": row.FormulaEnabled,
                "RawValue"      : row.RawValue,
                "RangedValue"   : row.RangedValue,
                "FormulaValue"  : analogInputs.getFormulaValue(row.Address, logger),
                "Units"         : analogInputs.getAnalogInputUnits(row.Address, logger),
                "ThresholdState": row.ThresholdState,
                "Enabled"       : row.Enabled,
                "SendEvents"    : row.SendEvents,
                "Range"         : row.Range,
                "LowThreshold"  : row.LowThreshold,
                "HighThreshold" : row.HighThreshold,
                "FormulaLowThreshold"  : analogInputs.getFormulaLowThreshold(row.Address, logger),
                "FormulaHighThreshold" : analogInputs.getFormulaHighThreshold(row.Address, logger),
                "Offset"        : row.Offset,
                "HoldTime"      : row.HoldTime,
                "Qualifier"     : row.Qualifier,
                "QualifierValue": analogInputs.getAnalogInputQualifierValue(row.Address, logger),
                "TimeChanged"   : row.TimeChanged
            });
            logger.debug('Summit API: GET /api/analoginputs/' + req.params.analoginput + ' - Points=' + JSON.stringify(_analogInput));
            res.json(_analogInput);
        }
        else {
            res.json({ Error: "Undefined analoginput" });
            throw logger.error('Summit API: GET /api/analoginputs/' + req.params.analoginput + ' - Error (Select)=' + JSON.stringify(err));
        }
    });
})

// Update Analog Input point
app.put('/api/analoginputs/:analoginput', function (req, res, next) {
    //  Name        (TEXT)      <device name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  SendEvents  (TEXT)      ”TRUE”, “FALSE”
    //  Polarity    (TEXT)      ”Normally Closed”, “Normally Open”
    //  ActionType  (TEXT)      "Latched", "Pulse"
    logger.info('Summit API: PUT /api/analoginputs/' + req.params.analoginput + " data="+JSON.stringify(req.body));
    
    var update = "UPDATE analoginputs SET Name='" + req.body.Name + "', FormulaId='" + req.body.FormulaId + "', FormulaEnabled='" + req.body.FormulaEnabled + "', Enabled='" + req.body.Enabled + "', SendEvents='" + req.body.SendEvents + "', Range='" + req.body.Range + "', LowThreshold='" + req.body.LowThreshold + "', HighThreshold='" + req.body.HighThreshold + "', Offset='" + req.body.Offset + "', Qualifier='" + req.body.Qualifier + "', HoldTime='" + req.body.HoldTime + "' WHERE address='" + req.params.analoginput + "'";
    db.run(update, function (err) {
        if (err !== null) {
            logger.error('Summit API: PUT /api/analoginputs/' + req.params.analoginput + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ Error : 'Error updating analoginput' })
        }
        else {
            // Modify the memory array
            analogInputs.UpdateAnalogInputConfiguration(req.params.analoginput, req.body.Name, req.body.FormulaId, req.body.FormulaEnabled, req.body.RawValue, req.body.RangedValue, req.body.FormulaValue, req.body.ThresholdState, req.body.Enabled, req.body.SendEvents, req.body.Range, req.body.LowThreshold, req.body.HighThreshold, req.body.FormulaLowThreshold, req.body.FormulaHighThreshold, req.body.Offset, req.body.HoldTime, req.body.Qualifier, req.body.TimeChanged, io, logger);
            analogInputs.ForceNewRawValue(req.params.analoginput, logger);// Force a read, in case something has changed, for example, new thresholds
            
            // Send back the req as results if everything OK
            var AnalogInput = analogInputs.GetAnalogInputValues(req.params.analoginput, logger);
            logger.debug('Summit API: PUT /api/analoginputs/' + req.params.analoginput + ' - Point=' + JSON.stringify(AnalogInput));
            res.send(JSON.stringify(AnalogInput))
        }
    });
})

//#endregion Analog Inputs

//#region Formulas

// Get Formulas Array
app.get('/api/formulas', function (req, res, next) {
    logger.info('Summit API: GET /api/formulas');
    
    // Get Formulas Array
    analogInputs.GetFormulas(db, io, logger, function (FormulasArray) {
        logger.debug('Summit API: GET /api/formulas - Formulas=' + JSON.stringify(FormulasArray));
        res.json(FormulasArray);
    });
});

// Add Formulas
app.put('/api/formulas', function (req, res, next) {
    //  Name        (TEXT)      <name>
    //  Enabled     (TEXT)      <formula>
    //  Code        (TEXT)      <units>
    //  Decimals    (INTEGER)   <decimals>
    //  TimeChanged (TEXT)      <time changed>
    logger.debug('Summit API: PUT /api/formulas - Formula=' + JSON.stringify(req.body));
    
    // Add Formula via req parameters
    analogInputs.AddFormula(req, db, io, logger, function (FormulasArray) {
        logger.debug('Summit API: PUT /api/formulas  - Formulas=' + JSON.stringify(FormulasArray));
        res.json(FormulasArray);
    });
});

// Modify existing formula
app.put('/api/formulas/evaluate', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  Name        (TEXT)      <name>
    //  Enabled     (TEXT)      <formula>
    //  Code        (TEXT)      <units>
    //  Decimals    (INTEGER)   <decimals>
    //  TimeChanged (TEXT)      <time changed>
    logger.debug('Summit API: PUT /api/formulas/evaluate - Formula=' + JSON.stringify(req.body));
    
    // Evaluate an existing Formula via req params
    analogInputs.EvaluateFormula(req.body, logger, function (result) {
        logger.debug('Summit API: PUT /api/formulas/evaluate - result=' + JSON.stringify(result));
        res.json(result);
    });
})

// Validate existing formula
app.put('/api/formulas/validate', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  Name        (TEXT)      <name>
    //  Enabled     (TEXT)      <formula>
    //  Code        (TEXT)      <units>
    //  Decimals    (INTEGER)   <decimals>
    //  TimeChanged (TEXT)      <time changed>
    logger.debug('Summit API: PUT /api/formulas/validate - Formula=' + JSON.stringify(req.body));
    
    // Validate an existing Formula via req params
    analogInputs.ValidateFormula(req, db, io, logger, function (result) {
        logger.debug('Summit API: PUT /api/formulas/validate - result=' + JSON.stringify(result));
        res.json(result);// true if parsed, false if parsing error
    });
})

// Delete a formula
app.put('/api/formulas/delete', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  Name        (TEXT)      <name>
    //  Enabled     (TEXT)      <formula>
    //  Code        (TEXT)      <units>
    //  Decimals    (INTEGER)   <decimals>
    //  TimeChanged (TEXT)      <time changed>
    logger.debug('Summit API: PUT /api/formulas/delete - Formula=' + JSON.stringify(req.body));
    
    // Delete an existing Formula via req params
    analogInputs.DeleteFormula(req, db, io, logger, function (FormulasArray) {
        logger.debug('Summit API: PUT /api/formulas  - Formulas=' + JSON.stringify(FormulasArray));
        res.json(FormulasArray);
    });

})

// Modify existing formula
app.put('/api/formulas/:formula', function (req, res, next) {
    //  Id          (TEXT)      <Id of the host>
    //  Name        (TEXT)      <name>
    //  Enabled     (TEXT)      <formula>
    //  Code        (TEXT)      <units>
    //  Decimals    (INTEGER)   <decimals>
    //  TimeChanged (TEXT)      <time changed>
    logger.info('Summit API: PUT /api/formulas/:' + req.params.formula + " Formula=" + JSON.stringify(req.body));
    
    // Modify an existing Formula via req params
    analogInputs.ModifyFormula(req, db, io, logger, function (FormulasArray) {
        logger.debug('Summit API: PUT /api/formulas/:' + FormulasArray.Id +' - Formulas=' + JSON.stringify(FormulasArray));
        res.json(FormulasArray);
    });
})

//#endregion Formulas

//#region Analog Outputs

// Get Analog Output points
app.get('/api/analogoutputs', function (req, res, next) {
    logger.info('Summit API: GET /api/analogoutputs');
    
    var _analogOutputs = [];
    db.all("SELECT Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged FROM analogoutputs ORDER BY Address", function (err, rows) {
        if (err) {
            throw logger.error('Summit API: GET /api/analogoutputs - Error (Select)=' + JSON.stringify(err));
        }
        
        for (i = 0; i < rows.length; i++) {
            _analogOutputs.push({
                "Address"       : rows[i].Address,
                "Name"          : rows[i].Name,
                "RawValue"      : rows[i].RawValue,
                "RangedValue"   : rows[i].RangedValue,
                "Enabled"       : rows[i].Enabled,
                "SendEvents"    : rows[i].SendEvents,
                "Range"         : rows[i].Range,
                "TimeChanged"   : rows[i].TimeChanged
            });
        }
        logger.debug('Summit API: GET /api/analogoutputs - Points=' + JSON.stringify(_analogOutputs));
        res.json(_analogOutputs);
    });
})

// Get Analog Output point
app.get('/api/analogoutputs/:analogoutput', function (req, res, next) {
    logger.info('Summit API: GET /api/analogoutputs/' + req.params.analogoutput);
    
    var _analogOutput = [];
    db.get("SELECT Address, Name, RawValue, RangedValue, Enabled, SendEvents, Range, TimeChanged FROM analogoutputs WHERE Address='" + req.params.analogoutput + "'", function (err, row) {
        if (err) {
            throw logger.error('Summit API: GET /api/analogoutputs/' + req.params.analogoutput + ' - Error (Select)=' + JSON.stringify(err));
        }
        
        if (typeof row !== "undefined") {
            _analogOutput.push({
                "Address"       : row.Address,
                "Name"          : row.Name,
                "RawValue"      : row.RawValue,
                "RangedValue"   : row.RangedValue,
                "Enabled"       : row.Enabled,
                "SendEvents"    : row.SendEvents,
                "Range"         : row.Range,
                "TimeChanged"   : row.TimeChanged
            });
            logger.debug('Summit API: GET /api/analogoutputs/' + req.params.analogoutput + ' - Point=' + JSON.stringify(_analogOutput));
            res.json(_analogOutput)
        }
        else {
            logger.error('Summit API: GET /api/analogoutputs/' + req.params.analogoutput + ' - Error (Select)=' + JSON.stringify(err));
            res.json({ Error: "Undefined analogoutput" });
        }
    });
})

// Update Analog Output point
app.put('/api/analogoutputs/:analogoutput', function (req, res, next) {
    //  Name        (TEXT)      <device name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  SendEvents  (TEXT)      ”TRUE”, “FALSE”
    //  Range       (TEXT)      <Voltage range>
    logger.info('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput);
    
    var update = "UPDATE analogoutputs SET Name='" + req.body.Name + "', Enabled='" + req.body.Enabled + "', SendEvents='" + req.body.SendEvents + "', Range='" + req.body.Range + "' WHERE address='" + req.params.analogoutput + "'";
    db.run(update, function (err) {
        if (err !== null) {
            logger.error('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ Error : 'Error updating analogoutput' })
        }
        else {
            if (req.body.Enabled == "FALSE") {
                var analogOutputRange = req.body.Range;
                var oldRangedValue = req.body.RangedValue;// Old Value
                req.body.RangedValue = "0.00";// Set to 0.00V because we are now disabled
                var analogOutput = parseInt(req.params.analogoutput);
                analogOutputRawValue = parseInt((req.body.RangedValue * ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE) + ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET);
                if (analogOutputRawValue < 0) analogOutputRawValue = 0;
                req.body.RawValue = analogOutputRawValue;
                logger.info('Summit API: PUT /api/analogoutputs/' + analogOutput + ' Previous RangedValue="'+ oldRangedValue+'" Now Ranged Value="'+ req.body.RangedValue+'" Converted Raw Value=' + analogOutputRawValue);
                
                fs.writeFile("/sys/class/hwmon/hwmon2/out" + analogOutput + "_output", analogOutputRawValue, function (err) {
                    if (err) {
                        logger.error('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + ' set RangedValue=0V - Error (Write)=' + JSON.stringify(err));
                        res.send({ Error : 'Error setting analogoutput' })
                    } else {
                        logger.debug('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + ' set RangedValue=0V - success');
                    }
                });
                
            }
            
            // Update the memory array
            analogOutputs.UpdateAnalogOutputConfiguration(req.params.analogoutput, req.body.Name, req.body.RawValue, req.body.RangedValue, req.body.Enabled, req.body.SendEvents, req.body.Range, req.body.HoldTime, io, logger)
            
            // Send back the req as results if everything OK
            logger.debug('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + ' - Point=' + JSON.stringify(req.body));
            res.send(JSON.stringify(req.body))
        }
    });
})

// Set Analog Output point value
var ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE = 200;
var ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET = 0;
app.put('/api/analogoutputs/:analogoutput/:value', function (req, res, next) {
    logger.info('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + '/' + req.params.value);
    
    if (req.params.value) {
        //Get range from the database, so we know what value to write out to the file
        var analogOutput = parseInt(req.params.analogoutput);
        var analogOutputRangedValue = req.params.value;
        var analogOutputRawValue = 0;
        
        // Get analog output point information       
        db.get("SELECT Range FROM AnalogOutputs WHERE Address='" + analogOutput + "'", function (err, row) {
            if (err) {
                logger.error('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + '/' + req.params.value + ' - Error (Select)=' + JSON.stringify(err));
                res.send({ Error : 'Error setting analogoutput' })
            }
            
            if (typeof row !== "undefined") {
                var analogOutputRange = row.Range;
                analogOutputRawValue = parseInt((analogOutputRangedValue * ANALOG_OUTPUT_CONVERT_TO_RAW_SLOPE) + ANALOG_OUTPUT_CONVERT_TO_RAW_OFFSET);
                if (analogOutputRawValue < 0) analogOutputRawValue = 0;
                logger.info('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + '/' + req.params.value + " Converted Raw Value=" + analogOutputRawValue);
                
                fs.writeFile("/sys/class/hwmon/hwmon2/out" + analogOutput + "_output", analogOutputRawValue, function (err) {
                    if (err) {
                        logger.error('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + '/' + req.params.value + ' - Error (Write)=' + JSON.stringify(err));
                        res.send({ Error : 'Error setting analogoutput' })
                    }
                });
                
                logger.debug('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + '/' + req.params.value + ' - Point=' + JSON.stringify(req.params));
                res.send(req.params)
            }
        });
    }
    else {
        logger.error('Summit API: PUT /api/analogoutputs/' + req.params.analogoutput + '/' + req.params.value + ' - Error (Value)');
        res.send({ Error : 'Error analogoutput value' })
    }
})

//#endregion Analog Outputs

//#region Serial Ports

//#region Device Options methods
// Get Device Options
app.get('/api/serialports/deviceoptions', function (req, res, next) {
    logger.info('Summit API: GET /api/serialports/deviceoptions');

    var serialPortDeviceOptions = SerialPortsManager.GetSerialPortDeviceOptions();
    logger.info('Summit API: GET /api/serialports/deviceoptions - DeviceOptions=' + JSON.stringify(serialPortDeviceOptions));
    res.send(JSON.stringify(serialPortDeviceOptions));
})
//#endregion Device Options methods

//#region Device Model methods
// Get Device Model
app.get('/api/serialports/devicemodel/:serialport', function (req, res, next) {
    logger.info('Summit API: GET /api/serialports/devicemodel/'+req.params.serialport);

    var deviceModel = SerialPortsManager.GetSerialPortDeviceModel(req.params.serialport);
    logger.info('Summit API: GET /api/serialports/devicemodel/' + req.params.serialport + ' - DeviceModel=' + JSON.stringify(deviceModel));
    res.send(JSON.stringify(deviceModel));
})

// Evaluate Formula
var Parser = require('expr-eval').Parser;// Library to manage parsing and evaluation of formulas
var mathparser = new Parser();
app.put('/api/serialports/devicemodel/formulas', function (req, res, next) {
    logger.info('Summit API: PUT /api/serialports/devicemode/formulas');

    var _expression
    for (var i = 0; i < req.body.Formulas.length; i++) {
        _expression = mathparser.parse(req.body.Formulas[i].Formula);
        req.body.Formulas[i].FormulaLowThreshold = Number(_expression.evaluate({ x: req.body.Formulas[0].FormulaLowThreshold })).toFixed(req.body.Formulas[i].Decimals);
        req.body.Formulas[i].FormulaHighThreshold = Number(_expression.evaluate({ x: req.body.Formulas[0].FormulaHighThreshold })).toFixed(req.body.Formulas[i].Decimals);
        logger.info('Summit API: PUT /api/serialports/devicemode/formulas - Low[' + i + ']=' + req.body.Formulas[i].FormulaLowThreshold + ' High[' + i + ']=' + req.body.Formulas[i].FormulaHighThreshold);
    }
    res.send(JSON.stringify(req.body));
})

// Put Device Model
app.put('/api/serialports/devicemodel/:serialport', function (req, res, next) {
    logger.info('Summit API: PUT /api/serialports/devicemodel/' + req.params.serialport);

    SerialPortsManager.UpdateSerialPortDeviceModel(req.params.serialport, req.body, function (devicemodel) {
        if (devicemodel === "Error") {
            logger.error('Summit API: PUT /api/serialports/devicemodel/' + req.params.serialport + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ Error: 'Error updating serialports/devicemodel' + req.params.serialport});
        } else {
            logger.debug('Summit API: PUT /api/serialports/devicemodel/' + req.params.serialport + ' - Point=' + JSON.stringify(devicemodel));
            res.send(JSON.stringify(devicemodel));
        }
    });
})

//#endregion Device Model methods

//#region Serial Port methods
// Get Serial Ports
app.get('/api/serialports', function (req, res, next) {
    logger.info('Summit API: GET /api/serialports');

    var SerialPortSettings = [];
    db.all("SELECT Address, Name, Enabled, SendEvents, Device, TimeChanged FROM serialports ORDER BY Address", function (error, rows) {
        if (error) {
            throw logger.error('Summit API: GET /api/serialports - Error (Select)=' + JSON.stringify(error));
        }

        for (i = 0; i < rows.length; i++) {
            SerialPortSettings.push({
                "Address": rows[i].Address,
                "Name": rows[i].Name,
                "Enabled": rows[i].Enabled,
                "SendEvents": rows[i].SendEvents,
                "Device": rows[i].Device,
                "TimeChanged": rows[i].TimeChanged
            });
        }
        logger.debug('Summit API: GET /api/serialports - Ports=' + JSON.stringify(SerialPortSettings));
        res.json(SerialPortSettings);
    });
})

// Get Serial Port x
app.get('/api/serialports/:serialport', function (req, res, next) {
    logger.info('Summit API: GET /api/serialports/' + req.params.serialport);

    var SerialPortSettings = [];
    db.get("SELECT Address, Name, Enabled, SendEvents, Device, TimeChanged FROM serialports WHERE Address=" + req.params.serialport, function (error, row) {
        if (error) {
            throw logger.error('Summit API: GET /api/serialports/' + req.params.serialport + ' - Error (Select)=' + JSON.stringify(error));
        }

        logger.debug('Summit API: GET /api/serialports/' + req.params.serialport + ' - row=' + JSON.stringify(row));

        if (typeof row !== "undefined") {
            // Always send the raw value - clients responsible for conversion
            logger.info('Summit API: GET /api/serialports/' + req.params.serialport + ' - Ports=' + JSON.stringify(row));
            res.json(row);
        } else {
            logger.error('Summit API: GET /api/serialports/' + req.params.serialport + ' - No Data');
            res.json({ Error: 'Error retrieving serialport' });
        }
    });
})

// Update Serial Port 
app.put('/api/serialports/:serialport', function (req, res, next) {
    //  Name        (TEXT)      <device name>
    //  Enabled     (TEXT)      ”TRUE”,”FALSE”
    //  Device      (TEXT)      ”NONE", <DeviceName>
    logger.info('Summit API: PUT /api/serialports/' + req.params.serialport);

    SerialPortsManager.UpdateSerialPortSettingPort(req.body, function (serialport) {
        if (serialport === "Error") {
            logger.error('Summit API: PUT /api/serialports/' + req.params.serialport + ' - Error (Update)=' + JSON.stringify(err));
            res.send({ Error: 'Error updating serialports' });
        } else {
            logger.debug('Summit API: PUT /api/serialports/' + req.params.serialport + ' - Point=' + JSON.stringify(serialport));
            res.send(JSON.stringify(serialport));
        }
    });
})

// Reset Serial Port x
app.put('/api/serialports/reset/:serialport', function (req, res, next) {
    logger.info('Summit API: PUT /api/serialports/reset/' + req.params.serialport);

    SerialPortsManager.ResetSerialPortSettingPort(req.params.serialport, function (result) {
        if (result > 0) {
            logger.error('Summit API: PUT /api/serialports/reset/' + req.params.serialport + ' - Error=' + JSON.stringify(result));
            res.send({ Status: 'ERROR' });
        } else {
            logger.info('Summit API: PUT /api/serialports/reset/' + req.params.serialport + ' - Success='+result);
            res.send({ Status: "OK" });
        }
    });

})

// Zero Serial Port x
app.put('/api/serialports/zero/:serialport', function (req, res, next) {
    logger.info('Summit API: PUT /api/serialports/zero/' + req.params.serialport);

    SerialPortsManager.ZeroSerialPortSettingPort(req.params.serialport, function (result) {
        if (result > 0) {
            logger.error('Summit API: PUT /api/serialports/zero/' + req.params.serialport + ' - Error=' + JSON.stringify(result));
            res.send({ Status: 'ERROR' });
        } else {
            logger.info('Summit API: PUT /api/serialports/zero/' + req.params.serialport + ' - Success=' + result);
            res.send({ Status: "OK" });
        }
    });

})
//#endregion Serial Port methods

//#endregion Serial Ports

//#region Kenwood Control

// Get Kenwood ping interval
app.get('/api/kenwood/pinginterval', function (req, res, next) {
    logger.info('Summit API: GET /api/kenwood/pinginterval');
    
    // Get the Kenwood ping interval
    logger.debug('Summit API: GET /api/kenwood/pinginterval - Kenwood Ping Interval=' + JSON.stringify(KENWOOD_PING_INTERVAL));
    res.json(KENWOOD_PING_INTERVAL);
});

// Set Kenwood ping interval
app.put('/api/kenwood/pinginterval', function (req, res, next) {
    logger.info('Summit API: PUT /api/kenwood/pinginterval');
    
    // Sets the Kenwood ping interval
    logger.debug('Summit API: PUT /api/kenwood/pinginterval - New ping interval=' + req.body.Interval);
    KENWOOD_PING_INTERVAL = req.body.Interval;
    res.json({ Interval : KENWOOD_PING_INTERVAL });
});

// Get Kenwood rules
app.get('/api/kenwood/rules', function (req, res, next) {
    logger.info('Summit API: GET /api/kenwood/rules');
    
    // Get the complete list of Kenwood rules
    kenwoodManager.GetKenwoodRules(logger, function (kenwoodRules) {
        logger.debug('Summit API: GET /api/kenwood/rules - Kenwood Rules=' + JSON.stringify(kenwoodRules));
        res.json(kenwoodRules);
    });
});

// Create new Kenwood rule
app.put('/api/kenwood/rules', function (req, res, next) {
    logger.info('Summit API: PUT /api/kenwood/rules');
    
    // Add a new Kenwood rule, 
    //  returns list of Kenwood rules( with new rule)
    kenwoodManager.AddKenwoodRule(req, logger, db, function (kenwoodRules) {
        logger.debug('Summit API: PUT /api/kenwood/rules - Kenwood Rules=' + JSON.stringify(kenwoodRules));
        res.json(kenwoodRules);
    });
});


// Delete a Kenwood rule
app.put('/api/kenwood/rules/delete', function (req, res, next) {
    logger.info('Summit API: PUT /api/kenwood/rules/delete');
    
    // Add a new Kenwood rule, 
    //  returns list of Kenwood rules( with new rule)
    kenwoodManager.DeleteKenwoodRule(req, logger, db, function (kenwoodRules) {
        logger.debug('Summit API: PUT /api/kenwood/rules/delete - Kenwood Rules=' + JSON.stringify(kenwoodRules));
        res.json(kenwoodRules);
    });
});
    
// Modify a Kenwood rule
app.put('/api/kenwood/rules/:kenwoodrule', function (req, res, next) {
    logger.info('Summit API: PUT /api/kenwood/rules/' + req.params.kenwoodrule);
    
    // Modidy an existing Kenwood rule, 
    //  returns list of Kenwood rules( with edited rule)
    kenwoodManager.ModifyKenwoodRule(req, logger, io, db, netSnmp, function (kenwoodRules) {
        logger.debug('Summit API: PUT /api/kenwood/rules=' + req.params.kenwoodrule + ' - Kenwood Rules=' + JSON.stringify(kenwoodRules));
        res.json(kenwoodRules);
    });
});
//#endregion Kenwood Control

//#region System Clock
//logger.info('Intializing System Clock')
//setInterval(sendTime, 1000)
//function sendTime() {
//    var currentTime = moment();
//    io.sockets.emit("CurrentSystemTime", { "CurrentSystemTime": currentTime });
//}
//#endregion System Clock

//#region Client Connections
server.listen(SUMMIT_WEB_SERVER_PORT, function () {
    logger.info('Web server listening on port ' + SUMMIT_WEB_SERVER_PORT)
});
//#endregion Client Connections

//---------------------------------------------
// Upgrade the database if required
//#region Database Upgrade

// Crest (Summit) upgrade strategy is as follows:
//  1. Check previous.json to determine previous version
//      1. a) If previous.json doesn't exists, then we're dealing with v0.13.xx or before
//  2. Determine the differences between package.json and previous.json and upgrade accordingly

fs.stat('previous.json', function (err, stat) {

    if (err == null) {
        // File exists, therefore this is for versions v0.14.xx and afterwards
        //  * Future upgrade code goes here *
        var previousjson = require('./previous.json');
        logger.info("Crest.DatabaseUpdate: previous.json exists - last installed version was v" + previousjson.version);
        logger.info("Crest.DatabaseUpdate: previous.json exists - new installed version is v" + pjson.version);
    
    } else if (err.code == "ENOENT") {
        // File doesn't exist, therefore we are dealing with a v0.13.xx or before, 
        //  so upgrade accordinglingly to v0.14.xx

        // v0.13.xx -> v0.14.xx
        //  - Add FormulaId column to the database
        logger.info("Crest.DatabaseUpdate: v0.13.xx -> v" + pjson.version + " Upgrade starting ...");
        
        // Check if the FormulaId column already exists
        db.all("PRAGMA table_info(AnalogInputs)", function (err, row) {
            if (err) {
                logger.error("Crest.DatabaseUpdate: v0.13.xx -> v" + pjson.version + " Upgrade - Error getting AnalogInputs table_info err=" + JSON.stringify(err));            
            } else if (row) {
                //logger.info("Crest.DatabaseUpdate: v0.13.xx -> v"+pjson.version+" Upgrade - Success getting AnalogInputs table_info=" + JSON.stringify(row));
                var FormulaIdColumnExists = false;
                for (var i = 0; i < row.length; i++) {
                    if (row[i].name == "FormulaId") {
                        FormulaIdColumnExists = true;
                        break;
                    }
                }
                
                // The FormulaId column doesn't exist, add it ...
                if (!FormulaIdColumnExists) {
                    logger.info("Crest.DatabaseUpdate: v0.13.xx -> v" + pjson.version + " Upgrade - FormulaId Column doesn't exist, adding it");
                    db.exec("ALTER TABLE AnalogInputs ADD COLUMN FormulaId INTEGER", function (err) {
                        if (err !== null) {
                            logger.error('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - ALTER TABLE AnalogInputs ADD COLUMN FormulaId err=' + JSON.stringify(err));
                        } else {
                            db.exec('UPDATE AnalogInputs SET FormulaId=-1', function (err) {
                                if (err !== null) {
                                    logger.error('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - "UPDATE AnalogInputs SET FormulaId=-1" err=' + JSON.stringify(err));
                                } else {
                                    logger.info('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - "UPDATE AnalogInputs SET FormulaId=-1" was successful');
                                    
                                    // Check if the FormulaEnabled column exists
                                    var FormulaEnabledColumnExists = false;
                                    for (var i = 0; i < row.length; i++) {
                                        if (row[i].name == "FormulaEnabled") {
                                            FormulaEnabledColumnExists = true;
                                            break;
                                        }
                                    }
                                    
                                    // If FormulaEnabled doesn't exist, create column and fill it with FormulaEnabled=1
                                    if (!FormulaEnabledColumnExists) {
                                        logger.info("Crest.DatabaseUpdate: v0.13.xx -> v" + pjson.version + " Upgrade - FormulaEnabled Column doesn't exist, adding it");
                                        db.exec("ALTER TABLE AnalogInputs ADD COLUMN FormulaEnabled INTEGER", function (err) {
                                            if (err !== null) {
                                                logger.error('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - ALTER TABLE AnalogInputs ADD COLUMN FormulaEnabled err=' + JSON.stringify(err));
                                            } else {
                                                db.exec('UPDATE AnalogInputs SET FormulaEnabled=1', function (err) {
                                                    if (err !== null) {
                                                        logger.error('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - "UPDATE AnalogInputs SET FormulaEnabled=1" err=' + JSON.stringify(err));
                                                    } else {
                                                        logger.info('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - "UPDATE AnalogInputs SET FormulaEnabled=1" was successful');
                                                    }
                                                });
                                            }
                                        });
                                    } else {
                                        logger.info('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - FormulaEnabled column already exists!');
                                    }
                                }
                            });
                        }
                    });
                } else {
                    logger.info('Crest.DatabaseUpgrade: v0.13.xx -> v' + pjson.version + ' - FormulaId column already exists!');
                }
            }
        });
    
    } else {
        logger.error("Crest.DatabaseUpdate: Error getting previous.json=" + JSON.stringify(err));
    }    
});

//#endregion Database Upgrade

//---------------------------------------------
// System Core
//#region System Core
var SnmpHostsArray = [];// Global array of SNMP Hosts
var KenwoodControlArray = [];// Global array of Kenwood Control Rules
var SnmpHostSessionsArray = [];// Global array of SNMP Host Sessions

setTimeout(function () {
    logger.info("Summit System Core starting up!")
    
    logger.info("System Core: Initialize the analog input formulas");
    analogInputs.InitializeFormulas(db, io, logger, function () {
        logger.info("System Core: Analog Input formulas are initialized");
    });
    
    logger.info("System Core: Starting up SNMP Hosts")
    networkManager.InitializeSnmpHostSettings(function (InitializedSnmpHostsArray) {
        // (Re)set the SNMP Host Array
        SnmpHostsArray = InitializedSnmpHostsArray.slice();
        logger.info("System Core: Initialized SnmpHostsArray=" + JSON.stringify(SnmpHostsArray));
        
        // (Re)set the SNMP Host Sessions for each I/O
        networkManager.CreateSnmpHostSessions(SnmpHostsArray, function (snmpHostSessions) {
            SnmpHostSessionsArray = snmpHostSessions.slice();
            logger.info("System Core: Initialized SnmpHostSessionsArray Length=" + SnmpHostSessionsArray.length);
            
            //---------------------------------------------
            //  Factory reset processing starts here....
            logger.info("System Core: Factory Reset Processing starting")
            
            factoryReset.ProcessFactoryReset(db, logger, io, netSnmp, systemStatusManager, function () {

            });
            
            //---------------------------------------------
            //  I/O processing starts here....
            //      We were waiting for SNMP Hosts to be initialized ...
            
            logger.info("System Core: I/O Processing starting")
            
            digitalInputs.InitializeDigitalInputs(smtpManager, db, io, logger, function () {
                logger.info('System Core IO: Digital Inputs initialized')
                digitalInputs.SetSnmpHostSessionsArray(SnmpHostSessionsArray, logger, function () {
                    logger.info('System Core IO: Digital Inputs being processed (' + DIGITAL_INPUTS_CYCLE_TIME + 'ms)')
                    setInterval(processDigitalInputs, DIGITAL_INPUTS_CYCLE_TIME);
                    function processDigitalInputs() {
                        digitalInputs.ProcessDigitalInputs(db, logger, io, netSnmp, systemStatusManager, function (digitalInputsArray) {
                            // After reading the digital input values, update the analog input qualifier values
                            analogInputs.UpdateQualifierValues(digitalInputsArray, db, logger, io)
                            SerialPortsManager.UpdateSerialPortQualifiers(digitalInputsArray);
                        })
                    }
                });
            });
            
            digitalOutputs.InitializeDigitalOutputs(smtpManager, db, io, logger, function () {
                logger.info('System Core IO: Digital Outputs initialized')
                digitalOutputs.ResetDigitalOutputs(db, logger, io, netSnmp, systemStatusManager, function () {
                    logger.info('System Core IO: Digital Outputs reset')
                    digitalOutputs.SetSnmpHostSessionsArray(SnmpHostSessionsArray, logger, function () {
                        logger.info('System Core IO: Digital Outputs being processed (' + DIGITAL_OUTPUTS_CYCLE_TIME + 'ms)')
                        setInterval(processDigitalOutputs, DIGITAL_OUTPUTS_CYCLE_TIME);
                        function processDigitalOutputs() {
                            digitalOutputs.ProcessDigitalOutputs(db, logger, io, netSnmp, systemStatusManager)
                        }
                        
                        // Since the Kenwood Manager is dependent on digital outputs, we need to have them started first
                        logger.info('System Core IO: Start Kenwood Manager');
                        kenwoodManager.InitializeKenwoodManager(db, io, logger, function (InitializedKenwoodControlArray) {
                            KenwoodControlArray = InitializedKenwoodControlArray.slice();
                            logger.info("System Core: Initialized KenwoodControlArray=" + JSON.stringify(KenwoodControlArray));
                            logger.info('System Core: Kenwood Host Pings being processed (' + KENWOOD_PING_INTERVAL * 1000 + 'ms)')
                            setTimeout(processKenwoodPing, KENWOOD_PING_INTERVAL * 1000);
                            function processKenwoodPing() {
                                kenwoodManager.ProcessKenwoodPings(db, logger, io, netSnmp, systemStatusManager, digitalOutputs, function () {
                                    // After pinging the host(s), do something if we need to ...
                                    //logger.info("SystemCore: Completed Processing of Kenwood Ping, reset Timeout to: "+ KENWOOD_PING_INTERVAL * 1000)
                                    setTimeout(processKenwoodPing, KENWOOD_PING_INTERVAL * 1000)
                                });
                            }
                        });
                    });
                });
            });
            
            analogInputs.InitializeAnalogInputs(smtpManager, db, io, logger, function () {
                logger.info('System Core IO: Analog Inputs initialized')
                analogInputs.SetSnmpHostSessionsArray(SnmpHostSessionsArray, logger, function () {
                    logger.info('System Core IO: Analog Inputs being processed (' + ANALOG_INPUTS_CYCLE_TIME + 'ms)')
                    setInterval(ProcessAnalogInputs, ANALOG_INPUTS_CYCLE_TIME);
                    function ProcessAnalogInputs() {
                        analogInputs.ProcessAnalogInputs(db, logger, io, netSnmp, systemStatusManager)
                    }
                });
            });
            
            analogOutputs.InitializeAnalogOutputs(smtpManager, db, io, logger, function () {
                logger.info('System Core IO: Analog Outputs initialized')
                analogOutputs.ResetAnalogOutputs(db, logger, io, netSnmp, systemStatusManager, function () {
                    logger.info('System Core IO: Analog Outputs reset')
                    analogOutputs.SetSnmpHostSessionsArray(SnmpHostSessionsArray, logger, function () {
                        logger.info('System Core IO: Analog Outputs being processed (' + ANALOG_OUTPUTS_CYCLE_TIME + 'ms)')
                        setInterval(processAnalogOutputs, ANALOG_OUTPUTS_CYCLE_TIME);
                        function processAnalogOutputs() {
                            analogOutputs.ProcessAnalogOutputs(db, logger, io, netSnmp, systemStatusManager)
                        }
                    });
                });
            });

            SerialPortsManager.InitializeSerialPorts(smtpManager, db, io, logger, netSnmp, function () {
                logger.info('System Core IO: Serial Ports initialized')
                SerialPortsManager.SetSnmpHostSessionsArray(SnmpHostSessionsArray, function () {
                    logger.info('System Core IO: Serial Ports being processed');
                });
            });

        });
    });
    
}, WAIT_UNTIL_SYSTEM_STARTS);
//#endregion System Core

